import ast
from dataclasses import dataclass, field
from typing import Dict, Optional

from transpiler.utils import highlight
from transpiler.phases.typing.annotations import TypeAnnotationVisitor
from transpiler.phases.typing.scope import Scope, ScopeKind, VarDecl
from transpiler.phases.typing.types import BaseType, TypeVariable, TY_NONE, TypeType
from transpiler.phases.utils import NodeVisitorSeq

PRELUDE = Scope.make_global()

@dataclass
class ScoperVisitor(NodeVisitorSeq):
    scope: Scope = field(default_factory=lambda: PRELUDE.child(ScopeKind.GLOBAL))
    root_decls: Dict[str, VarDecl] = field(default_factory=dict)
    cur_class: Optional[TypeType] = None

    def anno(self) -> "TypeAnnotationVisitor":
        return TypeAnnotationVisitor(self.scope, self.cur_class)

    def visit_annotation(self, expr: Optional[ast.expr]) -> BaseType:
        res = self.anno().visit(expr) if expr else TypeVariable()
        assert not isinstance(res, TypeType)
        return res

    def visit_block(self, block: list[ast.AST]):
        if not block:
            return
        TB = f"running type analysis on block starting with {highlight(block[0])}"
        self.fdecls = []
        for b in block:
            self.visit(b)
        if self.fdecls:
            old_list = self.fdecls
            exc = None
            while True:
                new_list = []
                for node, rtype in old_list:
                    from transpiler.exceptions import CompileError
                    try:
                        self.visit_function_definition(node, rtype)
                    except CompileError as e:
                        new_list.append((node, rtype))
                        if not exc or getattr(node, "is_main", False):
                            exc = e
                if len(new_list) == len(old_list):
                    raise exc
                if not new_list:
                    break
                old_list = new_list
                exc = None

    def visit_function_definition(self, node, rtype):
        TB = f"running type analysis on the body of {highlight(node)}"
        TB_NODE = node
        from transpiler.phases.typing.block import ScoperBlockVisitor
        for b in node.body:
            decls = {}
            visitor = ScoperBlockVisitor(node.inner_scope, decls)
            visitor.visit(b)
            b.decls = decls
        if not node.inner_scope.has_return:
            rtype.unify(TY_NONE)  # todo: properly indicate missing return

